@@ -5,8 +5,8 @@
 version_info = (
     1,
     2,
-    1,
-    # "",  # release (b1, rc1, or "" for final or dev)
+    2,
+    "",  # release (b1, rc1, or "" for final or dev)
     # "dev",  # dev or nothing for beta/rc/stable releases
 )
 
